import torch
from typing import Optional, Type, Union, Tuple
from torch import Tensor

class svk():
    """ Class for Saint Venant–Kirchhoff regularization.
    :param coef: weight of SVK
    :type coef: float
    :param lam: lamé constant
    :type lam: float
    :param mu: lamé constant
    :type mu: float
    :param pointsMask: a mask for pts if only a subset of pts needs to be measured
    :type pointsMask: torch.Tensor
    :param pointsMaskLabel: the mask label for pointsMask that needs to be measured
    :type pointsMaskLabel: int
    """
    def __init__(self,coef: float=1,lam: float=1,mu: float=1,pointsMask: Optional[Tensor] = None,pointsMaskLabel: Optional[Tensor] = None):
        """Constructor method."""
        self.lam=lam
        self.mu=mu
        self.coef=coef
        self.pointsMask=pointsMask
        self.pointsMaskLabel=pointsMaskLabel

    def __call__(self,dis_jac: Tensor=0,**kwargs)->Tensor:
        """Calculates SVK.
            :param dis_jac: jacobians of point displacements
            :type dis_jac: torch.Tensor
            :return: accumulated summed strainEnergy at time step
            :rtype: torch.Tensor"""
        #id = torch.eye(self.pts.size(1),device=self.pts.device).repeat((self.pts.size(0), 1, 1))
        #f=dis_jac+id
        if self.pointsMask is not None:
            dis_jac = dis_jac[self.pointsMask == self.pointsMaskLabel]
        else:
            pass
        jac = dis_jac
        jact = torch.transpose(jac, dim0=1, dim1=2)
        GreenTensor = 0.5 * (jac + jact + torch.einsum('abc,ade->abe', jact, jac))
        strainEnergy = (self.lam / 2 * torch.einsum('abb', GreenTensor) ** 2 + self.mu * torch.sum(GreenTensor ** 2,
                                                                                                   dim=[1, 2]))
        return self.coef*torch.sum(strainEnergy)



import torch
from typing import Optional, Type, Union, Tuple
from torch import Tensor

class nearest2D:
    """ Class for nearest neighbour interpolation in 2 dimensions.
    :param device: computation device, see torch docs
    :type device: str
    :param scale: tensor with stepsize between two consecutive data points in each dimension in pixel or voxel
    :type scale: torch.Tensor
    """
    def __init__(self,device,scale=None):
        """Constructor method.                """
        self.device=device
        self.scale=scale
    def __call__(self,pts,data):
        """ Compute nearest neighbour interpolation in 2 dimensions. Can interpolate on one field of data points or stacked fields
        :param pts: interpolation points coordinates
        :type pts: torch.Tensor
        :param data: tensor of data points (# fields,  dim 1,  dim 2 )
        :type data: torch.Tensor
        :return w: values at interpolation points (# fields,#points)
        :rtype w: torch.Tensor
        """
        if self.scale is None:
            self.scale=torch.ones(pts.size(1),device=self.device)
        t_idx = torch.round(pts.div(self.scale)).long()
        w=torch.zeros((data.size(0),pts.size(0))).to(device=self.device)
        t_idx[:, 0] = torch.clamp(t_idx[:, 0], 0, data.size(1)-1)
        t_idx[:, 1] = torch.clamp(t_idx[:, 1], 0, data.size(2)-1)
        for i,field in enumerate(data):
            w[i]=field[t_idx[:,0],t_idx[:,1]]
        return w


class nearest3D:
    """ Class for nearest neighbour interpolation in 3 dimensions.
    :param device: computation device, see torch docs
    :type device: str
    :param scale: tensor with stepsize between two consecutive data points in each dimension in pixel or voxel
    :type scale: torch.Tensor
    """
    def __init__(self,device: str,scale: Tensor)->Tensor:
        """Constructor method."""
        self.device=device
        self.scale=scale
    def __call__(self,pts,data):
        """ Compute nearest neighbour interpolation in 3 dimensions. Can interpolate on one field of data points or stacked fields
        :param pts: interpolation points coordinates
        :type pts: torch.Tensor
        :param data: tensor of data points (# fields,  dim 1,  dim 2, dim 3 )
        :type data: torch.Tensor
        :return w: values at interpolation points (# fields,#points)
        :rtype w: torch.Tensor
        """
        t_idx = torch.round(pts.div(self.scale)).long()
        w=torch.zeros((data.size(0),pts.size(0))).to(device=self.device)
        t_idx[:, 0] = torch.clamp(t_idx[:, 0], 0, data.size(1)-1)
        t_idx[:, 1] = torch.clamp(t_idx[:, 1], 0, data.size(2)-1)
        t_idx[:, 2] = torch.clamp(t_idx[:, 2], 0, data.size(3)-1)
        for i,field in enumerate(data):
            w[i]=field[t_idx[:,0],t_idx[:,1],t_idx[:,2]]
        return w

class nearest:
    """ Class for nearest neighbour interpolation.
    :param device: computation device, see torch docs
    :type device: str
    :param scale: tensor with stepsize between two consecutive data points in each dimension in pixel or voxel
    :type scale: torch.Tensor
    """
    def __new__(cls, device: str, scale: Tensor)->Union[nearest2D,nearest3D]:
        """Static method. Decides on dimensionality.
        :return: instance of interpolation class
        :rtype: nearest2D or nearest3D class"""
        if scale.size(0) == 2:
            return nearest2D(device, scale)
        elif scale.size(0) == 3:
            return nearest3D(device, scale)
        else:
            raise ValueError("Unsupported dimension. Only 2D and 3D are supported.")


import torch
import numpy as np
import os
from .. import interpolation, geometricTransformations,  numericalIntegration
import re
import matplotlib.pyplot as plt
import pickle
from .postprocessing_utils import *
from typing import Optional, Type, Union, Tuple
from torch import Tensor

def measure_dice(device: str,workingDirectory: str,segmentsOfInterest: Optional[list]=None)->Tensor:
    """ Calculates the Dice-Sørensen index based on segmentation masks for pullback images.
        :param device: sets the computation device, see torch
        :type device: str
        :param workingDirectory: path to working directory
        :type workingDirectory: str
        :param segmentsOfInterest: segmentation integers that are to be measured
        :type segmentsOfInterest: list
        :return: DICE
        :rtype: Tensor
    """
    movingImageMask=torch.unsqueeze(torch.from_numpy(np.load(os.path.join(workingDirectory, "moving_mask.npy"))), dim=0).to(device=device)
    fixedImageMask=torch.unsqueeze(torch.from_numpy(np.load(os.path.join(workingDirectory, "fixed_mask.npy"))), dim=0).to(device=device)
    indices = np.indices(movingImageMask.cpu()[0].size())
    pts = np.empty((np.prod(movingImageMask.cpu().size()), len(movingImageMask[0].cpu().size())))
    for i, slide in enumerate(indices):
        pts[:, i] = slide.flatten()
    pts=torch.from_numpy(pts).to(device=device).float()
    pts_orig=pts.clone()
    if segmentsOfInterest is not None:
        inter = interpolation.nearest(device,
                                      scale=torch.ones(pts.size(1), device=device))
        ptsSegmentation = inter(pts, movingImageMask).flatten().long()

        mask = torch.zeros_like(ptsSegmentation, dtype=torch.bool)
        for segment in segmentsOfInterest:
            mask |= (ptsSegmentation == segment)
        pts=pts[mask]
        ptsSegmentation=ptsSegmentation[mask]
        pts=checkAffine(device,workingDirectory,pts,ptsSegmentation)
        pts=checkNonrigid(device,workingDirectory,pts,ptsSegmentation)
        p=pts_orig.clone()
        p[mask]=pts
        pts=p
    else:
        pts=checkAffine(device,workingDirectory,pts)
        pts=checkNonrigid(device,workingDirectory,pts)
    intensityInterpolator = interpolation.nearest(device, torch.ones(pts.size(1), device=device))
    intensities = intensityInterpolator(pts, fixedImageMask)
    pullbackMask=intensities.flatten().reshape(fixedImageMask.size())
    overlap=torch.where(pullbackMask==movingImageMask,1.,0.)
    dice=2*(torch.sum(overlap))/(torch.numel(movingImageMask)+torch.numel(fixedImageMask))
    return dice











from .visual_convergence import *
from .visual_vector import *
from .measure_landmarks import *
from .visual_grid import *
from .measure_dice import *
from .measure_inverseConsistency import *
from .measure_selfintersection import *
from .measure_shear import *
from .visual_pullback import *
from .measure_intersection import *
from .measure_jacobian import *
from .visual_shear import *


import os.path
import igl
import torch
import numpy as np
import skfmm
from typing import Optional, Type, Union, Tuple
from torch import Tensor

class sdfCreator:
    """Class for creating signed distance fields (SDF).
    Voxels or pixels equating to maskLabel in mask are considered inside an object.
    :param device: computation device, see torch docs
    :type device: str
        :param reuse: whether the SDF should be reused
        :type reuse: bool
        :param workingDirectory: directory of the registration data
        :type workingDirectory: str
        :param segmentName: the label of the object to SDF is for
        :type segmentName: int
        """

    def __init__(self, device: str, reuse: Optional[bool]=False, workingDirectory: Optional[bool]=None, segmentName: Optional[int]=None):
        "Constructor method."
        self.segmentName = segmentName
        self.reuse = reuse
        self.workingDirectory = workingDirectory
        self.device = device

    def checkReuse(self)->bool:
        """Checks if SDF should be reused.
        If so, checks the reuse folder.
        :return: answer whether SDF was found or not
        :rtype: bool"""
        self.checkReuseFolder()
        print("reuse sdf")
        if os.path.exists(os.path.join(self.workingDirectory, "reuse", f"sdf_segment_{self.segmentName}.npy")):
            print("found")
            return True
        else:
            print(" not found")
            return False

    def loadReuse(self)->Tensor:
        """Reuses old SDF.
        :return: SDF
        :rtype: torch.Tensor"""
        return torch.from_numpy(
            np.load(os.path.join(self.workingDirectory, "reuse", f"sdf_segment_{self.segmentName}.npy"))).to(
            device=self.device)

    def saveReuse(self, sdf):
        """Saves SDF for reuse."""
        np.save(os.path.join(self.workingDirectory, "reuse", f"sdf_segment_{self.segmentName}.npy"),
                sdf.cpu().numpy())

    def fromMask(self, mask: Tensor, voxelSizes: Optional[Tensor]=None, invert: Optional[bool]=False)->Tensor:
        """Creates a SDF from a masked image. Entries with 1 will be assigned outside, everything else inside.
        :param mask: mask of the moving image (1,dim1,dim2(,dim3))
        :type mask: torch.Tensor
        :param voxelSizes: cell dimensions in mm
        :type voxelSizes: torch.Tensor
        :param invert: flip ones and non-ones integers in mask
        :type invert: bool"""
        if voxelSizes is None:
            if len(mask.size())==3:
                voxelSizes=torch.tensor([1.,1.])
            else:
                voxelSizes=torch.tensor([1.,1.,1.])
        else:
            pass
        if self.reuse:
            if self.checkReuse():
                return self.loadReuse()
            else:
                if invert:
                    mask = torch.where(mask == 1, -1., 1.)
                else:
                    mask = torch.where(mask == 1, 1., -1.)
                sdf = skfmm.distance(mask.cpu()[0], dx=voxelSizes.cpu())
                sdf = np.where(sdf > 0., 0, sdf)
                sdf = torch.unsqueeze(torch.from_numpy(sdf).to(device=self.device),dim=0)
                self.saveReuse(sdf)
                return sdf
        else:
            if invert:
                mask = torch.where(mask == 1, -1., 1.)
            else:
                mask = torch.where(mask == 1, 1., -1.)
            sdf = skfmm.distance(mask.cpu()[0], dx=voxelSizes.cpu())
            sdf = np.where(sdf > 0., 0, sdf)
            sdf = torch.unsqueeze(torch.from_numpy(sdf).to(device=self.device), dim=0)
            return sdf

    def fromMesh(self, pathToMesh: str, domain: Tensor)->Tensor:
        """Create a SDF from a 3D triangle surface mesh (stl files).
        :param pathToMesh: path to file
        :type pathToMesh: str
        :param domain: image domain (1,dim1,dim2,dim3)
        :type domain: torch.Tensor
        l"""
        vertices, faces = igl.read_triangle_mesh(pathToMesh, 'float')
        if self.reuse:
            if self.checkReuse():
                return self.loadReuse()
            else:
                indices = np.indices(domain.cpu()[0].size())
                pts = np.empty((np.prod(domain.cpu().size()), len(domain[0].cpu().size())))
                for i, slide in enumerate(indices):
                    pts[:, i] = slide.flatten()
                s, i, c = igl.signed_distance(pts, vertices, faces)
                ptsInt = pts.astype(int)
                sdf = domain.cpu().numpy().copy() * 0
                sdf[0,ptsInt[:, 0], ptsInt[:, 1], ptsInt[:, 2]] = s
                sdf = torch.from_numpy(sdf).to(device=self.device)
                self.saveReuse(sdf)
                return sdf
        else:
            indices = np.indices(domain.cpu()[0].size())
            pts = np.empty((np.prod(domain.cpu().size()), len(domain[0].cpu().size())))
            for i, slide in enumerate(indices):
                pts[:, i] = slide.flatten()
            s, i, c = igl.signed_distance(pts, vertices, faces)
            ptsInt = pts.astype(int)
            sdf = domain.cpu().numpy().copy() * 0
            sdf[0, ptsInt[:, 0], ptsInt[:, 1], ptsInt[:, 2]] = s
            sdf = torch.from_numpy(sdf).to(device=self.device)
            return sdf

    def checkReuseFolder(self):
        """Checks if a reuse folder is already created. If not, creates it."""
        if os.path.exists(os.path.join(self.workingDirectory, 'reuse/')):
            pass
        else:
            os.mkdir(os.path.join(self.workingDirectory, 'reuse/'))


    def getGrads(self, sdf: Tensor)->Tuple[Tensor,Tensor,Optional[Tensor]]:
        """Returns the gradients of SDF.
            :param sdf: signed distance field
            :type sdf: torch.Tensor
            :return gradX,gradY,gradZ: gradients
            :rtype sdf: list of torch.Tensor"""
        if len(sdf.size()) == 4:

            translated = sdf.cpu().numpy()[0]
            gradX, gradY, gradZ = np.gradient(translated)
            gradX = torch.unsqueeze(torch.from_numpy(gradX).to(device=sdf.device), dim=0)
            gradY = torch.unsqueeze(torch.from_numpy(gradY).to(device=sdf.device), dim=0)
            gradZ = torch.unsqueeze(torch.from_numpy(gradZ).to(device=sdf.device), dim=0)
            return [gradX, gradY, gradZ]

        elif len(sdf.size()) == 3:
            translated = sdf.cpu().numpy()[0]
            gradX, gradY = np.gradient(translated)
            gradX = torch.unsqueeze(torch.from_numpy(gradX).to(device=sdf.device), dim=0)
            gradY = torch.unsqueeze(torch.from_numpy(gradY).to(device=sdf.device), dim=0)
            return [gradX, gradY]
        else:
            raise Exception("check dimensions sdf")



import itertools

import skimage.measure
from shapely.geometry import Polygon
from skimage import measure
import torch
import numpy as np
import os
from ..import interpolation, geometricTransformations,  numericalIntegration, utils
import re
import matplotlib.pyplot as plt
import igl
import pickle
from .postprocessing_utils import *
from typing import Optional, Type, Union, Tuple
from torch import Tensor
def measure_intersection3d(device: str,workingDirectory: str,voxelToMm: Optional[Tensor]=None,segmentsOfInterest: Optional[list]=None,vertexLists: list=[],simplexLists: list=[])->Tuple[Tensor,Tensor]:
    """ Calculates the overlap and gap between segmentations in 3D. Objects are input as vertex and simplex collections.
        :param device: sets the computation device, see torch
        :type device: string
        :param workingDirectory: path to working directory, see docs
        :type workingDirectory: string
        :param segmentsOfInterest: segmentations of interest list
        :type segmentsOfInterest: list
        :param voxelToMm: cell dimensions in mm
        :type voxelToMm: torch.Tensor
        :param vertexLists: list of vertex collections
        :type vertexLists: list
        :param simplexLists: list of simplex collections
        :type simplexLists: list
        :return : gap and overlap
        :rtype : Tensor,Tensor
    """
    movingImage = torch.unsqueeze(torch.from_numpy(np.load(os.path.join(workingDirectory, "moving.npy"))), dim=0).to(
        device=device)
    indices = np.indices(movingImage.cpu()[0].size())
    coords = np.empty((np.prod(movingImage.cpu().size()), len(movingImage[0].cpu().size())))
    for i, slide in enumerate(indices):
        coords[:, i] = slide.flatten()
    coords = torch.from_numpy(coords).to(device=device).float()

    intersectionmap = torch.ones(movingImage[0].size(), device=device) * -1

    for i,segment in enumerate(segmentsOfInterest):
        vertices=vertexLists[i]
        simplices=simplexLists[i]
        ptsSegmentation = torch.ones(vertices.size(0), dtype=torch.bool)*segment
        vertices = checkAffine(device, workingDirectory, vertices, ptsSegmentation)
        vertices = checkNonrigid(device, workingDirectory, vertices, ptsSegmentation)

        p = igl.signed_distance(coords.cpu().numpy(), vertices.cpu().numpy(), simplices.long().cpu().numpy())
        sdf = movingImage[0].clone().float().to(device=device)
        sdf[coords[:, 0].long(), coords[:, 1].long(), coords[:, 2].long()] = torch.from_numpy(p[0]).flatten().to(
            device=device)
        sdf = torch.where(sdf < 0, 1, 0)
        intersectionmap = intersectionmap + sdf
    if voxelToMm is not None:
        return torch.sum(torch.where(intersectionmap == -1, 1, 0)) * voxelToMm[0] * voxelToMm[1] * voxelToMm[
        2], torch.sum(torch.where(intersectionmap >= 1, 1, 0)) * voxelToMm[0] * voxelToMm[1] * voxelToMm[2]
    else:
        return torch.sum(torch.where(intersectionmap == -1, 1, 0)) , torch.sum(torch.where(intersectionmap >= 1, 1, 0))

def measure_intersection2d(device: str,workingDirectory: str,segmentsOfInterest: Optional[list]=None,contours: list=[])->Tensor:
    """ Calculates the overlap between segmentations in 3D. Objects are input as contour collections.
        :param device: sets the computation device, see torch
        :type device: string
        :param workingDirectory: path to working directory, see docs
        :type workingDirectory: string
        :param segmentsOfInterest: segmentations of interest list
        :type segmentsOfInterest: list
        :param contours: contour collections
        :type contours: list
        :return : overlap
        :rtype : Tensor
    """
    movingImageMask=torch.unsqueeze(torch.from_numpy(np.load(os.path.join(workingDirectory, "moving_mask.npy"))), dim=0).to(device=device)
    indices = np.indices(movingImageMask.cpu()[0].size())
    coords = np.empty((np.prod(movingImageMask.cpu().size()), len(movingImageMask[0].cpu().size())))
    for i, slide in enumerate(indices):
        coords[:, i] = slide.flatten()
    contours=[]
    for i,segment in enumerate(segmentsOfInterest):
        contour=skimage.measure.find_contours(torch.where(movingImageMask[0]==segment,1,0).cpu().numpy(),level=0.5)[0]
        contour=torch.from_numpy(contour).to(device=device)
        ptsSegmentation = torch.ones(contour.size(0), dtype=torch.bool) * segment
        contour = checkAffine(device, workingDirectory, contour, ptsSegmentation)
        contour = checkNonrigid(device, workingDirectory, contour, ptsSegmentation)
        contours.append(contour)
    contoursIndices = np.arange(len(contours))
    combinations = list(itertools.combinations(contoursIndices, 2))
    dictIntersection={}
    for c in combinations:
        polygon1 = Polygon(contours[c[0]].cpu().numpy())
        polygon2 = Polygon(contours[c[1]].cpu().numpy())
        intersection = polygon1.intersection(polygon2)
        intersection_area = intersection.area
        dictIntersection[str(c)]=intersection_area
    return dictIntersection













import torch
import numpy as np
import os
from .. import interpolation, geometricTransformations,  numericalIntegration
import re
import matplotlib.pyplot as plt
import pickle
from .postprocessing_utils import *
from typing import Optional, Type, Union, Tuple
from torch import Tensor
def visual_pullback(device: str,workingDirectory: str,voxelToMm: Optional[list]=None,segmentsOfInterest: Optional[list]=None,dimension: int=None,slice: int=None):
    """ Plots the pullback image.
        :param device: sets the computation device, see torch
        :type device: string
        :param workingDirectory: path to working directory, see docs
        :type workingDirectory: string
        :param segmentsOfInterest: segmentations of interest list
        :type segmentsOfInterest: list
        :param voxelSizes: cell dimensions in mm
        :type voxelSizes: torch.Tensor
        :param dimension: if 3D registration, set with dimension should be displayed
        :type dimension: int
        :param slice: if 3D registration, set with slice in dimension should be displayed
        :type slice: int
    """

    #BASICS: load images
    movingImageMask=torch.unsqueeze(torch.from_numpy(np.load(os.path.join(workingDirectory, "moving_mask.npy"))), dim=0).to(device=device)
    movingImage=torch.unsqueeze(torch.from_numpy(np.load(os.path.join(workingDirectory, "moving.npy"))), dim=0).to(device=device)
    fixedImage=torch.unsqueeze(torch.from_numpy(np.load(os.path.join(workingDirectory, "fixed.npy"))), dim=0).to(device=device)

    indices = np.indices(movingImage.cpu()[0].size())
    pts = np.empty((np.prod(movingImage.cpu().size()), len(movingImage[0].cpu().size())))
    for i, slide in enumerate(indices):
        pts[:, i] = slide.flatten()
    pts=torch.from_numpy(pts).to(device=device).float()
    pts_orig=pts.clone()
    if segmentsOfInterest is not None:
        inter = interpolation.nearest(device,
                                      scale=torch.ones(pts.size(1), device=device))
        ptsSegmentation = inter(pts, movingImageMask).flatten().long()

        mask = torch.zeros_like(ptsSegmentation, dtype=torch.bool)
        for segment in segmentsOfInterest:
            mask |= (ptsSegmentation == segment)
        pts=pts[mask]
        ptsSegmentation=ptsSegmentation[mask]
        pts=checkAffine(device,workingDirectory,pts,ptsSegmentation)
        pts=checkNonrigid(device,workingDirectory,pts,ptsSegmentation)
        pts_orig=pts_orig[mask]
    else:
        pts = checkAffine(device, workingDirectory, pts)
        pts = checkNonrigid(device, workingDirectory, pts)

    interpolator=interpolation.cubic(device,torch.ones(pts.size(1),device=device))
    intensitiesMoving,_,_=interpolator(pts_orig,movingImage)
    movingImageResampled=movingImage.clone().float()
    intensitiesFixed,_,_=interpolator(pts_orig,fixedImage)
    fixedImageResampled=fixedImage.clone().float()
    pullback=movingImage.clone()*0.
    intensities,_,_=interpolator(pts,fixedImage)


    if dimension is not None and slice is not None:
        pullback[0,pts_orig[:,0].long(),pts_orig[:,1].long(),pts_orig[:,2].long()]=intensities.flatten()
        movingImageResampled[0,pts_orig[:,0].long(),pts_orig[:,1].long(),pts_orig[:,2].long()]=intensitiesMoving.flatten()
        fixedImageResampled[0,pts_orig[:,0].long(),pts_orig[:,1].long(),pts_orig[:,2].long()]=intensitiesFixed.flatten()

        if dimension==0:
            pullback=pullback[0,slice,:,:]
            movingImageResampled=movingImageResampled[0,slice,:,:]
            fixedImageResampled=fixedImageResampled[0,slice,:,:]
            extent = [0, movingImageResampled.size(1) * voxelToMm[2].cpu(), 0, movingImageResampled.size(0)  * voxelToMm[1].cpu()]

        elif dimension==1:
            pullback=pullback[0,:,slice,:]
            movingImageResampled=movingImageResampled[0,:,slice,:]
            fixedImageResampled=fixedImageResampled[0,:,slice,:]
            extent = [0, movingImageResampled.size(1) * voxelToMm[2].cpu(), 0, movingImageResampled.size(0) * voxelToMm[0].cpu()]


        elif dimension==2:
            pullback=pullback[0,:,:,slice]
            movingImageResampled=movingImageResampled[0,:,:,slice]
            fixedImageResampled=fixedImageResampled[0,:,:,slice]
            extent = [0, movingImageResampled.size(1) * voxelToMm[1].cpu(), 0, movingImageResampled.size(0) * voxelToMm[0].cpu()]


        else:
            raise Exception("postprocessing check dimensions")

        fig,ax=plt.subplots(4)
        ax[0].imshow(movingImageResampled.cpu(),extent=extent,origin='lower')
        ax[0].set_title("movingImageResampled")
        ax[1].imshow(pullback.cpu(),extent=extent,origin='lower')
        ax[1].set_title("pullback")
        ax[2].imshow((movingImageResampled-fixedImageResampled).cpu(),extent=extent,origin='lower')
        ax[2].set_title("movingImageResampled-fixedImageResampled")
        ax[3].imshow((movingImageResampled-pullback).cpu(),extent=extent,origin='lower')
        ax[3].set_title("movingImageResampled-pullback")
        plt.show()
    else:
        pullback[0,pts_orig[:,0].long(),pts_orig[:,1].long()]=intensities.flatten()
        movingImageResampled[0,pts_orig[:,0].long(),pts_orig[:,1].long()]=intensitiesMoving.flatten().float()
        fixedImageResampled[0,pts_orig[:,0].long(),pts_orig[:,1].long()]=intensitiesFixed.flatten().float()
        extent = [0, movingImage.size(2) * voxelToMm[1].cpu(), 0, movingImage.size(1) * voxelToMm[0].cpu()]
        fig,ax=plt.subplots(4)
        ax[0].imshow(movingImageResampled[0].cpu(),extent=extent,origin='lower')
        ax[0].set_title("movingImageResampled")
        ax[1].imshow(pullback[0].cpu(),extent=extent,origin='lower')
        ax[1].set_title("pullback")
        ax[2].imshow((movingImageResampled-fixedImageResampled)[0].cpu(),extent=extent,origin='lower')
        ax[2].set_title("movingImageResampled-fixedImageResampled")
        ax[3].imshow((movingImageResampled-pullback)[0].cpu(),extent=extent,origin='lower')
        ax[3].set_title("movingImageResampled-pullback")
        plt.show()
        plt.imshow((movingImageResampled-fixedImageResampled)[0].cpu()-(movingImageResampled-pullback)[0].cpu())
        plt.show()








import torch
import numpy as np
import os
from .. import interpolation, geometricTransformations,  numericalIntegration
import re
import matplotlib.pyplot as plt
import pickle
from .postprocessing_utils import *
from typing import Optional, Type, Union, Tuple
from torch import Tensor


def measure_inverseConsistency(device: str,workingDirectory: str,voxelToMm: Optional[list]=None,segmentsOfInterest: Optional[list]=None)->Tuple[Tensor,Tensor]:
    """ Calculates the inverse consistency.
        :param device: sets the computation device, see torch
        :type device: string
        :param workingDirectory: path to working directory, see docs
        :type workingDirectory: string
        :param segmentsOfInterest: segmentations of interest list
        :type segmentsOfInterest: list
        :param voxelSizes: cell dimensions in mm
        :type voxelSizes: torch.Tensor
        :return : mean and standard deviation of inverse consistency
        :rtype : Tensor,Tensor
    """

    #BASICS: load images
    movingImageMask=torch.unsqueeze(torch.from_numpy(np.load(os.path.join(workingDirectory, "moving_mask.npy"))), dim=0).to(device=device)
    movingImage=torch.unsqueeze(torch.from_numpy(np.load(os.path.join(workingDirectory, "moving.npy"))), dim=0).to(device=device)
    indices = np.indices(movingImage.cpu()[0].size())
    pts = np.empty((np.prod(movingImage.cpu().size()), len(movingImage[0].cpu().size())))
    for i, slide in enumerate(indices):
        pts[:, i] = slide.flatten()
    pts=torch.from_numpy(pts).to(device=device).float()
    pts_orig=pts.clone()

    if segmentsOfInterest is not None:
        inter = interpolation.nearest(device,
                                      scale=torch.ones(pts.size(1), device=device))
        ptsSegmentation = inter(pts, movingImageMask).flatten().long()

        mask = torch.zeros_like(ptsSegmentation, dtype=torch.bool)
        for segment in segmentsOfInterest:
            mask |= (ptsSegmentation == segment)
        pts=pts[mask]
        ptsSegmentation=ptsSegmentation[mask]
        pts=checkNonrigid(device,workingDirectory,pts,ptsSegmentation)
        pts=checkNonrigidInverse(device,workingDirectory,pts,ptsSegmentation)
        pts_orig=pts_orig[mask]

    else:
        pts = checkAffine(device, workingDirectory, pts)
        pts = checkNonrigid(device, workingDirectory, pts)
        pts=checkNonrigidInverse(device,workingDirectory,pts)

    if pts.size(1)==3:
        dist=pts-pts_orig
        dist[:,0]=dist[:,0]*voxelToMm[0]
        dist[:,1]=dist[:,1]*voxelToMm[1]
        dist[:,2]=dist[:,2]*voxelToMm[2]
        dist=torch.norm(dist,dim=1)
    else:
        dist=pts-pts_orig
        dist[:,0]=dist[:,0]*voxelToMm[0]
        dist[:,1]=dist[:,1]*voxelToMm[1]
        dist=torch.norm(dist,dim=1)

    return torch.mean(dist),torch.std(dist),dist






import torch
from typing import Optional, Type, Union, Tuple
from torch import Tensor






class cubic3d:
    """ Class for cubic interpolation in 3 dimensions.
    :param device: computation device, see torch docs
    :type device: str
    :param scale: tensor with stepsize between two consecutive data points in each dimension in pixel or voxel
    :type scale: torch.Tensor
    :param jac: switch to compute jacobian
    :type jac: bool
    :param lap: switch to compute laplacian
    :type lap: bool
    """
    def __init__(self, device: str, scale: Tensor,jac: bool,lap: bool):
        """Constructor method.                """
        self.device=device
        self.scale=scale
        self.ones4 = torch.ones([4, 4, 4], dtype=torch.int32, device=device)
        self.stride_x, self.stride_y, self.stride_z = torch.meshgrid(
            [torch.arange(0, 4, device=device) - 1, torch.arange(0, 4, device=device) - 1,
             torch.arange(0, 4, device=device) - 1])
        self.jac=jac
        self.lap=lap
    def __call__(self,pts: Tensor,data: Tensor)->Tuple[Tensor,Tensor,Tensor]:
        """ Compute cubic approximation in 3 dimensions.
        :param pts: interpolation points coordinates
        :type pts: torch.Tensor
        :param data: tensor of data points (# fields,  dim 1,  dim 2, dim 3 )
        :type data: torch.Tensor
        :return w: values at interpolation points (# fields,#points)
        :rtype w: torch.Tensor
        :return jacobian: jacobian of interpolation values
        :rtype jacobian: torch.Tensor
        :return laplacian: laplacian of interpolation values
        :rtype laplacian: torch.Tensor
        """
        p = torch.arange(pts.size(0))

        t_idx = pts.mul(1 / self.scale).floor()
        t = pts.mul(1 / self.scale) - t_idx
        onesp = torch.ones(t.size(0), dtype=torch.int32, device=self.device)

        t_idx = t_idx.flatten()
        indices = torch.clamp(
            torch.einsum('a,bcd->abcd', t_idx[3 * p], self.ones4) + torch.einsum('a,bcd->abcd', onesp, self.stride_x.int()),
            0,
            data.shape[1] - 1) * (data.size(2) * data.size(3))
        indices = indices + torch.clamp(
            torch.einsum('a,bcd->abcd', t_idx[3 * p + 1], self.ones4) + torch.einsum('a,bcd->abcd', onesp,
                                                                                     self.stride_y.int()),
            0,
            data.shape[2] - 1) * (data.size(3))
        indices = indices + torch.clamp(
            torch.einsum('a,bcd->abcd', t_idx[3 * p + 2], self.ones4) + torch.einsum('a,bcd->abcd', onesp,self.stride_z.int()),
            0,
            data.shape[3] - 1)
        torch.cuda.empty_cache()
        a = torch.stack([t.flatten() * 0 + 1, t.flatten(), t.flatten() ** 2, t.flatten() ** 3], dim=1).float()
        b = torch.tensor(([1, 4, 1, 0], [-3, 0, 3, 0], [3, -6, 3, 0], [-1, 3, -3, 1]), dtype=torch.float32,
                         device=self.device) / 6
        y = torch.mm(a, b)
        w = torch.sum(
            torch.einsum('ab,ac,ad->abcd', y[3 * p, :], y[3 * p + 1, :], y[3 * p + 2, :]) * data.flatten(start_dim=1)[:,
                                                                                            indices.long()], dim=[2, 3, 4])
        if self.jac:
            da = torch.stack([t.flatten() * 0, t.flatten() * 0 + 1, t.flatten() * 2, 3 * t.flatten() ** 2], dim=1)
            dy = torch.mm(da, b)
            wx = torch.sum(
                torch.einsum('ab,ac,ad->abcd', dy[3 * p, :], y[3 * p + 1, :], y[3 * p + 2, :]) * data.flatten(start_dim=1)[
                                                                                                 :, indices.long()],
                dim=[2, 3, 4])
            wy = torch.sum(
                torch.einsum('ab,ac,ad->abcd', y[3 * p, :], dy[3 * p + 1, :], y[3 * p + 2, :]) * data.flatten(start_dim=1)[
                                                                                                 :, indices.long()],
                dim=[2, 3, 4])
            wz = torch.sum(
                torch.einsum('ab,ac,ad->abcd', y[3 * p, :], y[3 * p + 1, :], dy[3 * p + 2, :]) * data.flatten(start_dim=1)[
                                                                                                 :, indices.long()],
                dim=[2, 3, 4])
            jacobian = torch.stack([wx.t(), wy.t(), wz.t()], dim=2)
        else:
            jacobian=0
        if self.lap:
            dda = torch.stack([t.flatten() * 0, t.flatten() * 0, t.flatten() * 0 + 1, 6 * t.flatten()], dim=1)
            ddy = torch.mm(dda, b)
            wwx = torch.sum(
                torch.einsum('ab,ac,ad->abcd', ddy[3 * p, :], y[3 * p + 1, :], y[3 * p + 2, :]) * data.flatten(
                    start_dim=1)[
                                                                                                  :, indices.long()],
                dim=[2, 3, 4])
            wwy = torch.sum(
                torch.einsum('ab,ac,ad->abcd', y[3 * p, :], ddy[3 * p + 1, :], y[3 * p + 2, :]) * data.flatten(
                    start_dim=1)[
                                                                                                  :, indices.long()],
                dim=[2, 3, 4])
            wwz = torch.sum(
                torch.einsum('ab,ac,ad->abcd', y[3 * p, :], y[3 * p + 1, :], ddy[3 * p + 2, :]) * data.flatten(
                    start_dim=1)[
                                                                                                  :, indices.long()],
                dim=[2, 3, 4])
            laplacian = wwx + wwy + wwz
        else:
            laplacian=0
        return w,jacobian,laplacian




class cubic2d:
    """ Class for cubic interpolation in 2 dimensions.
    :param device: computation device, see torch docs
    :type device: str
    :param scale: tensor with stepsize between two consecutive data points in each dimension in pixel or voxel
    :type scale: torch.Tensor
    :param jac: switch to compute jacobian
    :type jac: bool
    :param lap: switch to compute laplacian
    :type lap: bool
    """
    def __init__(self,device: str,scale: Tensor,jac: bool,lap: bool):
        """Constructor method.             """
        self.device=device
        self.scale=scale
        self.ones4 = torch.ones([4, 4], dtype=torch.int32, device=device)
        self.stride_x, self.stride_y = torch.meshgrid(
            [torch.arange(0, 4, device=device) - 1, torch.arange(0, 4, device=device) - 1])
        self.jac = jac
        self.lap = lap

    def __call__(self,pts: Tensor,data: Tensor)->Tuple[Tensor,Tensor,Tensor]:
        """ Compute cubic approximation in 2 dimensions.
        :param pts: interpolation points coordinates
        :type pts: torch.Tensor
        :param data: tensor of data points (# fields,  dim 1,  dim 2 )
        :type data: torch.Tensor
        :return w: values at interpolation points (# fields,#points)
        :rtype w: torch.Tensor
        :return jacobian: jacobian of interpolation values
        :rtype jacobian: torch.Tensor
        :return laplacian: laplacian of interpolation values
        :rtype laplacian: torch.Tensor
        """
        p = torch.arange(pts.size(0))
        t_idx = pts.div(self.scale).floor()
        t = pts.mul(1 / self.scale) - t_idx
        onesp = torch.ones(t.size(0), dtype=torch.int32, device=self.device)
        t_idx = t_idx.flatten()
        indices = torch.clamp(
            torch.einsum('a,bc->abc', t_idx[2 * p], self.ones4) + torch.einsum('a,bc->abc', onesp, self.stride_x.int()), 0,
            data.shape[1] - 1).long() * (data.size(2))
        indices = indices + torch.clamp(
            torch.einsum('a,bc->abc', t_idx[2 * p + 1], self.ones4) + torch.einsum('a,bc->abc', onesp, self.stride_y.int()),
            0,
            data.shape[2] - 1).long()
        tf = t.flatten()
        a = torch.stack([tf * 0 + 1, tf, tf ** 2, tf ** 3], dim=1).float()
        b = torch.tensor(([1, 4, 1, 0], [-3, 0, 3, 0], [3, -6, 3, 0], [-1, 3, -3, 1]), dtype=torch.float32,
                         device=self.device) / 6
        y = torch.mm(a, b)
        w = torch.sum(
            torch.einsum('ab,ac->abc', y[2 * p, :], y[2 * p + 1, :]) * data.flatten(start_dim=1)[:, indices.long()],
            dim=[2, 3])
        if self.jac:
            da = torch.stack([t.flatten() * 0, t.flatten() * 0 + 1, t.flatten() * 2, 3 * t.flatten() ** 2], dim=1)
            dy = torch.mm(da, b)
            wx = torch.sum(
                torch.einsum('ab,ac->abc', dy[2 * p, :], y[2 * p + 1, :]) * data.flatten(start_dim=1)[
                                                                            :, indices.long()],
                dim=[2, 3])
            wy = torch.sum(
                torch.einsum('ab,ac->abc', y[2 * p, :], dy[2 * p + 1, :]) * data.flatten(start_dim=1)[
                                                                            :, indices.long()],
                dim=[2, 3])

            jacobian = torch.stack([wx.t(), wy.t()], dim=2)
        else:
            jacobian=0
        if self.lap:
            dda = torch.stack([t.flatten() * 0, t.flatten() * 0, t.flatten() * 0 + 1, 6 * t.flatten()], dim=1)
            ddy = torch.mm(dda, b)
            wwx = torch.sum(
                torch.einsum('ab,ac->abc', ddy[2 * p, :], y[2 * p + 1, :]) * data.flatten(start_dim=1)[:, indices.long()],
                dim=[2, 3])
            wwy = torch.sum(
                torch.einsum('ab,ac->abc', y[2 * p, :], ddy[2 * p + 1, :]) * data.flatten(start_dim=1)[:, indices.long()],
                dim=[2, 3])
            print(wwx.size())
            laplacian = wwx + wwy
        else:
            laplacian=0
        return w,jacobian,laplacian



class cubic:
    """ Class for cubic interpolation.
    :param device: computation device, see torch docs
    :type device: str
    :param scale: tensor with stepsize between two consecutive data points in each dimension in pixel or voxel
    :type scale: torch.Tensor
    :param jac: switch to compute jacobian
    :type jac: bool
    :param lap: switch to compute laplacian
    :type lap: bool
    """
    def __new__(cls, device: str, scale: Tensor,jac: Optional[bool] = False,lap: Optional[bool] = False)->Union[cubic2d,cubic3d]:
        """Static method. Decides on dimensionality.
        :return: instance of interpolation class
        :rtype: interpolation class"""
        if scale.size(0) == 2:
            return cubic2d(device, scale,jac,lap)
        elif scale.size(0) == 3:
            return cubic3d(device, scale,jac,lap)
        else:
            raise ValueError("Unsupported dimension. Only 2D and 3D are supported.")


import numpy as np
import os
import re
import matplotlib.pyplot as plt
import json
from .postprocessing_utils import *
from typing import Optional, Type, Union, Tuple
from torch import Tensor

def visual_convergence(workingDirectory: str,segmentsOfInterest: Optional[list]=None):
    """ Plots convergence.
        :param workingDirectory: path to working directory, see docs
        :type workingDirectory: string
         :param segmentations: segmentations of interest list
         :type segmentations: list

    """
    if os.path.exists(os.path.join(workingDirectory,"results", "transformation_affine.json")):
        with open(os.path.join(workingDirectory, "results", "transformation_affine.json")) as json_file:
            data = json.load(json_file)
        fig,ax=plt.subplots(len(data.keys()))
        for i,key in enumerate(data.keys()):
            ax[i].plot(np.array(data[key]))
            ax[i].set_title(key)
        plt.suptitle('affine')
        plt.show()
    else:
        pass
    if segmentsOfInterest is not None:
        for s in segmentsOfInterest:
            print(s)
            if os.path.exists(os.path.join(workingDirectory,"results", f"transformation_affine_{s}.json")):
                with open(os.path.join(workingDirectory, "results", f"transformation_affine_{s}.json")) as json_file:
                    data = json.load(json_file)
                fig, ax = plt.subplots(len(data.keys()))
                for i, key in enumerate(data.keys()):
                    ax[i].plot(np.array(data[key]))
                    ax[i].set_title(key)
                plt.suptitle(f'affine segment {s}')
                plt.show()

    files = os.listdir(os.path.join(workingDirectory, "results"))
    filtered_files = [file for file in files if file.startswith("transformation_nonrigid") and file.endswith(".json")]

    if len(filtered_files) > 0:
        sorted_files = sorted(filtered_files,
                              key=lambda s: (extract_segment_and_scale_JSON(s)[0], extract_segment_and_scale_JSON(s)[1]),
                              reverse=True)
        for f in sorted_files:
            scale,segment=extract_segment_and_scale_JSON(f)
            with open(os.path.join(workingDirectory, "results", f)) as json_file:
                data = json.load(json_file)
            fig, ax = plt.subplots(len(data.keys()))
            for i, key in enumerate(data.keys()):
                ax[i].plot(data[key])
                ax[i].set_title(key)
            plt.suptitle(f'non rigid; segment {segment}; scale {scale}')
            plt.show()






